import numpy as np
import scipy as sp
import pdb, sys, pickle
import matplotlib
import matplotlib.pylab as plt
import pandas
from Tango import *

matplotlib.rcParams.update({'xtick.labelsize':16})

def plot_GO(golist, models = ['LIMMI', 'SVA'], cutoff = 0.05):
    assert len(golist) <= 2, 'this function supports comparisons between two models only'
    
    tango = Tango()
    plt.figure()
    # plt.axes([0.1, 0.65, 0.85, 0.3])
    plt.axes([0.60, 0.1, 0.3, 0.85]) 
    if len(golist) == 2:
	cur_width = 0.0
	width = 0.4
    else:
	cur_width = 0.0
	width = 0.8

    model_counts = []
    for g in golist:
	go = pandas.read_csv(g, sep='\t', index_col = 1)
	counts = go[go['Benjamini'] <= cutoff]['Count']
	model_counts.append(counts)

    if len(golist) == 2:
	go_categories = list(set.union(set(model_counts[0].keys()),
				       set(model_counts[1].keys())))
    else:
	go_categories = model_counts[0].keys()

	
    for i in range(len(models)):
	labels, vals = [], []
	for c in go_categories:
	    labels.append(c.split('~')[1])
	    try:
		vals.append(model_counts[i][c])
	    except KeyError:
		vals.append(0.0)


	pos = sp.arange(len(labels))
 	plt.barh(pos + cur_width, vals, height = width,
 		align='edge', color = tango.nextMedium())
	cur_width += width

    plt.yticks(pos+0.4, labels)
    plt.xlabel('Number of genes')
    plt.grid(False)
    plt.savefig('test.pdf')
    return go


if __name__ == '__main__':
    go = plot_GO(['../../papers/LIMMI_paper/supp/GO_annot_LIMMI.txt'], models = ['LIMMI'])
    #'../papers/LIMMI_paper/supp/GO_annot_SVA.txt'])



import scipy as SP
import numpy as NP
import scipy.linalg as LA
import scipy.optimize as OPT
import scipy.stats as st
import pdb


# log of 2pi
L2pi = 1.8378770664093453
def nLLeval(ldelta,UY,UX,S,MLparams=False):
    """evaluate the negative LL of a LMM with kernel USU.T"""
    delta=SP.exp(ldelta);
    n,d=UX.shape;
    Sdi=S+delta;
    ldet=SP.sum(NP.log(Sdi));
    Sdi=1.0/Sdi;
    XSdi=UX.T*SP.tile(Sdi,(d,1));
    XSX=SP.dot(XSdi,UX);
    XSY=SP.dot(XSdi,UY);
    beta=LA.lstsq(XSX,XSY);
    res=UY-SP.dot(UX,beta[0]);
    res*=res;
    res*=Sdi;
    sigg2=SP.sum(res)/n;
    nLL=0.5*(n*L2pi+ldet+n+n*NP.log(sigg2));
    if MLparams:
        return nLL, beta[0], sigg2;
    else:
        return nLL;

def optdelta(UY,UX,S,ldeltanull=None,numintervals=100,ldeltamin=-10.0,ldeltamax=10.0):
    """find the optimal delta"""
    if ldeltanull==None:
        nllgrid=SP.ones(numintervals+1)*SP.inf;
        ldeltagrid=SP.arange(numintervals+1)/(numintervals*1.0)*(ldeltamax-ldeltamin)+ldeltamin;
        nllmin=SP.inf;
        for i in SP.arange(numintervals+1):
            nllgrid[i]=nLLeval(ldeltagrid[i],UY,UX,S);
            if nllgrid[i]<nllmin:
                nllmin=nllgrid[i];
                ldeltaopt_glob=ldeltagrid[i];
        foundMin=False
        for i in SP.arange(numintervals-1)+1:
            ee = 1E-8
            #carry out brent optimization within the interval
            if ((nllgrid[i-1]-nllgrid[i])>ee) and ((nllgrid[i+1]-nllgrid[i])>1E-8):
                foundMin = True
                ldeltaopt,nllopt,iter,funcalls = OPT.brent(nLLeval,(UY,UX,S),(ldeltagrid[i-1],ldeltagrid[i],ldeltagrid[i+1]),full_output=True);
                if nllopt<nllmin:
                    nllmin=nllopt;
                    ldeltaopt_glob=ldeltaopt;
    else:
        ldeltaopt_glob=ldeltanull;
    return ldeltaopt_glob;

def train(X,Y,K,covariates=None,addBiasTerm=False,numintervalsAlt=0,ldeltaminAlt=-1.0,ldeltamaxAlt=1.0,numintervals0=100,ldeltamin0=-5.0,ldeltamax0=5.0):
    """ compute all pvalues
    If numintervalsAlt==0 use EMMA-X trick (keep delta fixed over alternative models)
    """
    n,s=X.shape;
    n_pheno=Y.shape[1];
    S,U=LA.eigh(K);
    UY=SP.dot(U.T,Y);
    UX=SP.dot(U.T,X);
    if (covariates==None):
        Ucovariate=SP.dot(U.T,SP.ones([n,1]));
    else:
        if (addBiasTerm):
            covariates_=SP.concatenate((covariates,SP.ones([n,1])),axis=1)
            Ucovariate=SP.dot(U.T,covariates_);
        else:
            Ucovariate=SP.dot(U.T,covariates);
    n_covar=Ucovariate.shape[1];
    beta = SP.empty((n_pheno,s,n_covar+1));
    beta0 = SP.empty((n_pheno,n_covar));
    LL=SP.ones((n_pheno,s))*(-SP.inf);
    LL0=SP.ones((n_pheno))*(-SP.inf);
    ldelta=SP.empty((n_pheno,s));
    ldelta0=SP.empty(n_pheno);
    sigg2=SP.empty((n_pheno,s));
    sigg20=SP.empty((n_pheno));
    pval=SP.ones((n_pheno,s))*(-SP.inf);
    for phen in SP.arange(n_pheno):
        UY_=UY[:,phen];
        ldelta0[phen]=optdelta(UY_,Ucovariate,S,ldeltanull=None,numintervals=numintervals0,ldeltamin=ldeltamin0,ldeltamax=ldeltamax0);
        nLL0_, beta0_, sigg20_=nLLeval(ldelta0[phen],UY_,Ucovariate,S,MLparams=True);
        beta0[phen,:]=beta0_;
        sigg20[phen]=sigg20_;
        LL0[phen]=-nLL0_;
        for snp in SP.arange(s):
            UX_=SP.hstack((UX[:,snp:snp+1],Ucovariate));
            if numintervalsAlt==0: #EMMA-X trick #fast version, no refitting of detla
                ldelta[phen,snp]=ldelta0[phen];                
            else: #fit delta
                ldelta[phen,snp]=optdelta(UY_,UX_,S,ldeltanull=None,numintervals=numintervalsAlt,ldeltamin=ldelta0[phen]+ldeltaminAlt,ldeltamax=ldelta0[phen]+ldeltamaxAlt);
            nLL_, beta_, sigg2_=nLLeval(ldelta[phen,snp],UY_,UX_,S,MLparams=True);
            beta[phen,snp,:]=beta_;
            sigg2[phen,snp]=sigg2_;
            LL[phen,snp]=-nLL_;
    #reshaping of LL0
    LL0 = LL0[:,SP.newaxis]           
    pval = st.chi2.sf(2*(LL-LL0),1)
    return LL0, LL, pval, ldelta0, sigg20, beta0, ldelta, sigg2, beta



def train_interact(X,Y,K,interactants=None,covariates=None,addBiasTerm=True,numintervalsAlt=0,ldeltaminAlt=-1.0,ldeltamaxAlt=1.0,numintervals0=10,ldeltamin0=-5.0,ldeltamax0=5.0):
    """ compute all pvalues
    If numintervalsAlt==0 use EMMA-X trick (keep delta fixed over alternative models)
    difference to previous model: Ux and Ucovariate are recomputed for every SNP
    """
    n,s=X.shape;
    n_pheno=Y.shape[1];
    S,U=LA.eigh(K);
    UY=SP.dot(U.T,Y);
    UX=SP.dot(U.T,X);
    if (covariates==None):
        covariates = SP.ones([n,0])
    if (addBiasTerm):
        covariates=SP.concatenate((covariates,SP.ones([n,1])),axis=1)
    #Ucovariates
    Ucovariate=SP.dot(U.T,covariates);

    #Uinteractants
    Uinteractants = SP.dot(U.T,interactants)
    n_covar=covariates.shape[1]
    n_inter=interactants.shape[1]
    #weights
    #foreground: covaraits + SNP + interactions 
    beta = SP.empty((n_pheno,s,1+n_covar+2*n_inter));
    #background: covariates + direct SNP effect
    beta0 = SP.empty((n_pheno,s,1+n_covar+n_inter));
    LL=SP.ones((n_pheno,s))*(-SP.inf);
    LL0=SP.ones((n_pheno,s))*(-SP.inf);
    ldelta=SP.empty([n_pheno,s]);
    ldelta0=SP.empty([n_pheno,s]);
    sigg2=SP.empty((n_pheno,s));
    sigg20=SP.empty((n_pheno,s));
    pval=SP.ones((n_pheno,s))*(-SP.inf);
    for snp in SP.arange(s):
        #loop through all SNPs
        #1. snp-specific backgroud model SNP effect + covaraites + interactants
        Ucovariates_=SP.hstack((UX[:,snp:snp+1],Uinteractants,Ucovariate))
        #2. snp-specific foreground model
        #interactions
        Xi_ = X[:,snp:snp+1]*interactants
        #transform
        UXi_ = SP.dot(U.T,Xi_)
        #stack: interactions, interactants (main) SNPs (main) covariates (if any)
        UX_  = SP.hstack((UXi_,Ucovariates_))
        pdb.set_trace()
        for phen in SP.arange(n_pheno):
            print phen
            #loop through phenoptypes
            #get transformed Y
            UY_=UY[:,phen]        
            #1. fit background model
            ldelta0[phen,snp]=optdelta(UY_,Ucovariates_,S,ldeltanull=None,numintervals=numintervals0,ldeltamin=ldeltamin0,ldeltamax=ldeltamax0);
            nLL0_, beta0_, sigg20_=nLLeval(ldelta0[phen,snp],UY_,Ucovariates_,S,MLparams=True);
            beta0[phen,snp,:]=beta0_;
            sigg20[phen,snp]=sigg20_;
            LL0[phen,snp]=-nLL0_;            

            #2. fit foreground model
            if numintervalsAlt==0: #EMMA-X trick #fast version, no refitting of detla
                ldelta[phen,snp]=ldelta0[phen,snp]                
            else: #fit delta
                ldelta[phen,snp]=optdelta(UY_,UX_,S,ldeltanull=None,numintervals=numintervalsAlt,ldeltamin=ldelta0[phen,snp]+ldeltaminAlt,ldeltamax=ldelta0[phen,snp]+ldeltamaxAlt);
            nLL_, beta_, sigg2_=nLLeval(ldelta[phen,snp],UY_,UX_,S,MLparams=True);
            beta[phen,snp,:]=beta_;
            sigg2[phen,snp]=sigg2_;
            LL[phen,snp]=-nLL_;
    pval = st.chi2.sf(2*(LL-LL0),1)
    return LL0, LL, pval, ldelta0, sigg20, beta0, ldelta, sigg2, beta



def train_interactX(X,Y,K,interactants=None,covariates=None,addBiasTerm=True,numintervalsAlt=0,ldeltaminAlt=-1.0,ldeltamaxAlt=1.0,numintervals0=10,ldeltamin0=-5.0,ldeltamax0=5.0):
    """ compute all pvalues
    If numintervalsAlt==0 use EMMA-X trick (keep delta fixed over alternative models)
    difference to previous model: Ux and Ucovariate are recomputed for every SNP
    """
    n,s=X.shape;
    n_pheno=Y.shape[1];
    S,U=LA.eigh(K);
    UY=SP.dot(U.T,Y);
    UX=SP.dot(U.T,X);
    if (covariates==None):
        covariates = SP.ones([n,0])
    if (addBiasTerm):
        covariates=SP.concatenate((covariates,SP.ones([n,1])),axis=1)
    #Ucovariates
    Ucovariate=SP.dot(U.T,covariates);

    #Uinteractants
    Uinteractants = SP.dot(U.T,interactants)
    n_covar=covariates.shape[1]
    n_inter=interactants.shape[1]
    #weights
    #foreground: covaraits + SNP + interactions 
    beta = SP.empty((n_pheno,s,1+n_covar+2*n_inter));
    #background: covariates + direct SNP effect
    beta0 = SP.empty((n_pheno,s,1+n_covar+n_inter));
    LL=SP.ones((n_pheno,s))*(-SP.inf);
    LL0=SP.ones((n_pheno,s))*(-SP.inf);
    ldelta=SP.empty([n_pheno,s]);
    ldelta0=SP.empty([n_pheno,s]);
    sigg2=SP.empty((n_pheno,s));
    sigg20=SP.empty((n_pheno,s));
    pval=SP.ones((n_pheno,s))*(-SP.inf);
    #0. fit 0 model on phenotypes and covariates alone
    for phen in SP.arange(n_pheno):
        #fit if phen is visited the first time
        #loop through phenoptypes
        #get transformed Y
        UY_=UY[:,phen]        
        #1. fit background model to set delta
        ldelta0[phen,:]=optdelta(UY_,Ucovariate,S,ldeltanull=None,numintervals=numintervals0,ldeltamin=ldeltamin0,ldeltamax=ldeltamax0);
            
    #1. loop through all snps
    for snp in SP.arange(s):
        #loop through all SNPs
        #1. snp-specific backgroud model SNP effect + covaraites + interactants
        Ucovariates_=SP.hstack((UX[:,snp:snp+1],Uinteractants,Ucovariate))
        #2. snp-specific foreground model
        #interactions
        Xi_ = X[:,snp:snp+1]*interactants
        #transform
        UXi_ = SP.dot(U.T,Xi_)
        #stack: interactions, interactants (main) SNPs (main) covariates (if any)
        UX_  = SP.hstack((UXi_,Ucovariates_))

        for phen in SP.arange(n_pheno):
            UY_=UY[:,phen]        
            #loop through all phenotypes
            #emmaX trick
            ldelta[phen,snp]=ldelta0[phen,snp]
            #evluate background and foreground
            #null model
            nLL0_, beta0_, sigg20_=nLLeval(ldelta0[phen,snp],UY_,Ucovariates_,S,MLparams=True)
            beta0[phen,snp,:]=beta0_
            sigg20[phen,snp]=sigg20_
            LL0[phen,snp]=-nLL0_                   
            #foreground model
            nLL_, beta_, sigg2_=nLLeval(ldelta[phen,snp],UY_,UX_,S,MLparams=True)
            beta[phen,snp,:]=beta_
            sigg2[phen,snp]=sigg2_
            LL[phen,snp]=-nLL_
    pval = st.chi2.sf(2*(LL-LL0),1)
    return LL0, LL, pval, ldelta0, sigg20, beta0, ldelta, sigg2, beta


def run_interact(Y, intA, intB, covs, K):
    """ Calculate pvalues for the nested model of including a multiplicative term between intA and intB into the additive model """
    [N, Ny] = Y.shape

    Na = intA.shape[1] # number of interaction terms 1
    Nb = intB.shape[1] # number of interaction terms 2
    
    S,U=LA.eigh(K);
    UY=SP.dot(U.T,Y);
    UintA=SP.dot(U.T,intA);
    UintB=SP.dot(U.T,intB);
    Ucovs=SP.dot(U.T,covs);
    # for each snp/gene/factor combination, run a lod
    # snps need to be diced bc of missing values - iterate over them, else in arrays
    lods = SP.zeros([Na, Nb, Ny])

    #add mean column:
    if covs is None: covs = SP.ones([Ny,1])

    # for each pair of interacting terms
    for a in range(Na):
        for b in range(Nb):
            # calculate additive and interaction terms
            C = SP.concatenate((Ucovs, UintA[:,a:a+1], UintB[:,b:b+1]))
            X = intA[:,a:a+1]*intB[:,b:b+1]
            UX = SP.dot(U.T,X);
            UX = SP.concatenate((UX, C))
            for phen in SP.arange(Ny):
                UY_=UY[:,phen];
                nllnull,ldeltanull=optdelta(UY_,C,S,ldeltanull=None,numintervals=10,ldeltamin=-5.0,ldeltamax=5.0);
                nllalt,ldeltaalt=optdelta(UY_,UX,S,ldeltanull=ldeltanull,numintervals=100,ldeltamin=-5.0,ldeltamax=5.0);
                lods[a,b,phen] = nllalt-nllalt;
    return lods



"""
Package for Priors of Gaussian Processes
========================================

This package provides priors for gaussian processes in which
you can declare your prior beliefs of the hyperparameter
distribution.

"""

try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)



"""
Adapted version from pybedtools

"""
import pdb

def venn_mpl(a, b, c, colors=None, outfn=None, labels=None):
    """
    *a*, *b*, and *c* are filenames to BED-like files.

    *colors* is a list of matplotlib colors for the Venn diagram circles.

    *outfn* is the resulting output file.  This is passed directly to
    fig.savefig(), so you can supply extensions of .png, .pdf, or whatever your
    matplotlib installation supports.

    *labels* is a list of labels to use for each of the files; by default the
    labels are ['a','b','c']
    """

    import matplotlib.pyplot as plt
    from matplotlib.patches import Circle

    if colors is None:
        colors = ['r','b','g']

    radius = 6.0
    center = 0.0
    offset = radius / 2

    if labels is None:
        labels = ['a','b','c']

    circle_a = Circle(xy = (center-offset, center+offset), radius=radius, edgecolor=colors[0], label=labels[0])
    circle_b = Circle(xy = (center+offset, center+offset), radius=radius, edgecolor=colors[1], label=labels[1])
    circle_c = Circle(xy = (center,        center-offset), radius=radius, edgecolor=colors[2], label=labels[2])


    fig = plt.figure(facecolor='w')
    ax = fig.add_subplot(111)

    for circle in (circle_a, circle_b, circle_c):
        circle.set_facecolor('none')
        circle.set_linewidth(3)
        ax.add_patch(circle)

    ax.axis('tight')
    ax.axis('equal')
    ax.set_axis_off()


    kwargs = dict(horizontalalignment='center')

    # Unique to A
    ax.text( center-2*offset, center+offset, str(len(a - b - c)), **kwargs)

    # Unique to B
    ax.text( center+2*offset, center+offset, str(len(b - a - c)), **kwargs)

    # Unique to C
    ax.text( center, center-2*offset, str(len(c - a - b)), **kwargs)

    # A and B not C
    ax.text( center, center+2*offset-0.5*offset, str(len((a & b) - c)), **kwargs)

    # A and C not B
    ax.text( center-1.2*offset, center-0.5*offset, str(len((a & c) - b)), **kwargs)

    # B and C not A
    ax.text( center+1.2*offset, center-0.5*offset, str(len((b & c) - a)), **kwargs)

    # all
    ax.text( center, center, str(len(a & b & c)), **kwargs)

    ax.legend(loc='best')

    fig.savefig(outfn)
    plt.close(fig)


# venn_mpl(set([1,2,3]), set([1,3,5,6]), set([3,54,1,4]))



"""
GPasso
- Gaussian process association test
"""

import sys, os, string, pickle, copy, pdb
sys.path.append('./../../pygp')
sys.path.append('./../../')
#gplvm model
from pygp.covar import linear, noise, combinators, fixed, mu
import pygp.gp.gplvm as GPLVM
import pygp.gp.gplvm_ard as GPLVM_ARD
# customized covariance function:
from interaction_covar import *
import pygp.priors.lnpriors as lnpriors
import pygp.likelihood as lik
import linear_regression as linreg
import logging as LG
import scipy as SP
import numpy as N
from panama.utilities import fdr
import scipy.stats as st
import scipy.sparse as sparse
import testing, interaction_scan


class GPLVMassoJ(object):
    """some details:
    fixed parameters (setData)
    snp_dimensions        : dimension index of snps (0..Ns)
    factor_dimensions     : dimension index of factors (Ns..Ns+Nf)

    learning parameters
    candidate_assoications : index with SNPs to include in the factor learning
    candidate_interactions : index with factor x SNP pairs to include in the factor learning
    ard: perform ard on the factors, i.e. estimate one varinace per factor
    lam: L1 penalty for factors
    """

    def __init__(self, sigma = 0.1, n_factors = 100, gplvm_init = 'randn', model = 'gplvm', add_associations = True,
		 pop_cf = False, mean_cf = True, asso_method = 'LMM', interact_method ='LMM', keep_latent = False,
		 parallelize = False, max_int_lod = False, Njobs = 5, ard = False, lam = 0, verbose_plot_dir = None,
		 Kpop = None, FA_noise = False, FDR_addition_associations = 0.2, FDR_addition_interactions = 0.01,
		 topn_interactions = 1, sim = None, testing_directory = None, fixed_latent_dim = False, pv_max_stored = None):

        self.testing_directory = testing_directory
        self.sim=sim
        self.n_factors = n_factors
        self.gplvm_init = string.upper(gplvm_init)
        self.model  = model
        self.sigma  = sigma
        self.candidate_interactions = []
        self.candidate_associations = []
	self.interactions_qv = []
	self.interactions_pv = []	

	self.fixed_latent_dim = fixed_latent_dim
	
        #iterations
        self.train_iterations = 0
	self.interactions_iter = 0

	# threshold for storing pvalues (stores only pv < thresh)
	self.pv_max_stored = pv_max_stored
	
        # threshold for adding stuff both in normal PANAMA and in LIMMI
        self.FDR_addition_associations = FDR_addition_associations
        self.FDR_addition_interactions = FDR_addition_interactions
	self.topn_interactions = topn_interactions # only add the top n inter.
	
        self.add_associations = add_associations
        self.use_pop_cf = pop_cf
        self.use_mean_cf = mean_cf
        self.keep_latent = keep_latent
	self.asso_method = asso_method
        self.cutoff_factors = 0 # threshold that decides when a factor is off
	
	# Add_interactions will be switched on later if gplvm_interact is selected
	self.add_interactions = False
        self.hyperparams = None  # erase hyperparams
        self.lmls = [] # log marginal likelihoods
        self.parallelize = parallelize
        self.Njobs = Njobs
        self.max_int_lod = max_int_lod
        self.ard = ard # (Boolean) should we use the ARD covariance?
        self.lam = lam
        self.verbose_plot_dir = verbose_plot_dir
	self.Kpop = Kpop
	# use a different noise level for each dimension
	self.FA_noise = FA_noise

	
    def setData(self, snps, pheno, covariates = None, pop_snps = None,normalize_pheno=True,normalize_geno=True):
	"""
	Loads the data, centering and scaling the SNPs and the phenotype

	"""

        self.nN = snps.shape[0]
        self.nD = pheno.shape[1]
        self.nS = snps.shape[1]


        if covariates is None:
            covariates = SP.zeros([self.nN,0])

        self.nC = covariates.shape[1]

        assert snps.shape[0]==pheno.shape[0], 'dimension mismatch'
        assert covariates.shape[0]==snps.shape[0], 'dimension mismatch'

        #uncentered raw SNPS
        #there are needed for interaction tests
        self.snps_raw = snps.copy()

	# Normalize SNPs
        if normalize_geno:
            LG.info("Centering and scaling the SNPs...")
            snps = snps - snps.mean(axis=0)
            snps /= snps.std(axis=0)

        if normalize_pheno:
            # Normalize genes
            LG.info("Centering and scaling the phenotype...")
            pheno = pheno - pheno.mean(axis=0)
            pheno /= pheno.std(axis=0)

        # SNPS/PHENO
        self.snps  = snps
        self.pheno = pheno

        self.covariates = covariates

        #set starting SNP dimensions and factor dimensions
        self.pop_snps = pop_snps

        #initialize GP object
	self._init_gp()

    def lMl(self):
        """
	Returns the NEGATIVE log marginal likelihood

	"""

        return self.gp.LML(self.hyperparams)

    def setLatent(self,x):
        """
	Sets the latent variables
	"""
        self.hyperparams['x'] = x


    def findK(self,**kw_args):
        """
	Returns the number of PCA dimensions that explain 95% of the variance

	"""

    	Y = self.pheno.copy()
    	Y -= Y.mean(axis=0)
    	Y /= Y.std(axis=0)
    	U, s, V = SP.linalg.svd(Y, full_matrices = False)
    	eigv = SP.sort(s**2)[::-1]
    	expl = SP.cumsum(eigv)

    	return min(self.n_factors,SP.nonzero(expl/expl.max() > 0.90)[0].min())


    def get_covar_param_index(self, covar):
	"""
	Finds the index of the hyperparameters for a given covariance function

	"""

        # determine indices of covariance:
        icf = 0
        np = 0
        # find covariance parametrs
        while (self.covar_list[icf] != covar):
            np += self.covar_list[icf].get_number_of_parameters()
            icf +=1
        assert self.covar_list[icf] == covar, 'covariance not found!'

        nparam = covar.n_hyperparameters
        Ip = SP.arange(np,np+nparam)

        return Ip

    def get_explained_var(self):
	hyperparams = {'covar': copy.deepcopy(self.hyperparams['covar']),
		       'lik': copy.deepcopy(self.hyperparams['lik']),
		       'x': copy.deepcopy(self.hyperparams['x'])}
	
	for cf in self.covar_list:
	    cf_name = cf.__class__.__name__
	    param_index = self.get_covar_param_index(cf)
	    param = hyperparams['covar'][param_index]
	    if cf_name == "LinearCFARD":
		param = 1/param
		X_var = hyperparams['x'].var(axis=0)
		param = X_var * param
	    elif cf_name == "LinearCF":
		param = SP.exp(param)
		X_var = self.snps[:,self.candidate_associations].var(axis=0)
		param = X_var * param
	    elif cf_name == "LinearInteractCF":
		if len(self.candidate_interactions) > 0:
		    param = SP.exp(param)
		    X_var = hyperparams['x'].var(axis=0)[SP.array(self.candidate_interactions)[:,0]]
		    S_var = self.snps.var(axis=0)[SP.array(self.candidate_interactions)[:,1]]
		    param = param*S_var*X_var
	    else:
		param = SP.exp(param)

		
	    print cf_name, param.round(4), param.sum().round(4)

	# noise
	param = SP.exp(hyperparams['lik'])
	print "noise", param.round(4)

	
    def get_K_testing(self,correct_trans = True, correct_confounders = True,
		      correct_interactions = True, noise = True, Ix = None,
		      refit_confounders = True):

        """
	Returns a corrected covariance structure

        correct_trans: include genetic component (True)
        correct_confounders: include confounders in covariance (True)
        noise: include noise (True)
        """

        # 1. create copy of current hyperparmas, dropping latent factors X
	hyperparams = {'covar': copy.deepcopy(self.hyperparams['covar']),
		       'lik': copy.deepcopy(self.hyperparams['lik'])}
	
        I_snp = self.get_covar_param_index(self.snp_cf)
        # find confounder covariance parameters
        I_factor = self.get_covar_param_index(self.factor_cf)
        # filter for optimization
        Ifilter = {'covar': SP.ones(len(hyperparams['covar']),dtype='bool'),
		   'lik': SP.ones(len(hyperparams['lik']),dtype='bool')}

	# If we are in interaction mode, this option kills the interaction
	# covariance
	if self.model == "gplvm_interact" and not correct_interactions:
	    index_interactions = self.get_covar_param_index(self.interact_cf)
	    hyperparams['covar'][index_interactions] = SP.log(1E-10)
            # construct filter for optimizer         
            Ifilter['covar'][index_interactions] = False
	    
        if not correct_trans:
            # switch off SNP effects
            hyperparams['covar'][I_snp] = SP.log(1E-10)
            # construct filter for optimizer         
            Ifilter['covar'][I_snp] = False

        if not correct_confounders:
            # switch off confounder CF
            # note these are not! operated in log space and are 1/p!
            hyperparams['covar'][I_factor] = 1E10
            Ifilter['covar'][I_factor] = False

	if not refit_confounders:
	    # correct_confounders kills the confounders and takes them out of the optimization
	    # this statement just removes them from the opt
            Ifilter['covar'][I_factor] = False

        # reoptimize covariance structure:
        # REMEMBER: excludes latent variables from optimization
        [hyperparams_o,opt_lml_o] = GPLVM.opt_hyper(self.gp,hyperparams,Ifilter=Ifilter,bounds=self.hyperparams_bounds,priors = self.hyperparams_priors)

        if Ix is not None:
	    # switch off some of the factors
	    factors_off = I_factor[~Ix]
            hyperparams_o['covar'][factors_off] = 1E10

        if not noise:
            # disables noise
	    if self.FA_noise:
		hyperparams_o['lik'] = SP.log(1E-10)
	    else:
		hyperparams_o['lik'] = SP.log(SP.ones(self.nD)*1E-10)

        # compute the covariance matrix
        K   = self.gp.covar.K(hyperparams_o['covar'],self.gp.x)
	
        return [K,hyperparams_o]

    def _get_residuals(self, prediction = False, **kw_args):
        """calculate PANAMA residuals"""
        #1. get K without or with trans guys:
        [K,hyperparams_o] = self.get_K_testing(noise=True,**kw_args)
        KV = self.gp.get_covariances(hyperparams_o)
        #2. get cross covariance (merely removing noise here)
        Kstar = self.gp.covar.K(hyperparams_o['covar'],self.gp.x,self.gp.x)
        #3. calc predictions manually:
        mu = SP.dot(Kstar.transpose(),KV['alpha'])
        Y_ = self.pheno-mu
	if prediction:
	    return Y_, mu
	else:
	    return Y_

    def getAssociations(self,correct_trans=False):
	"""
        Association scan given current training state of the PANAMA model
	Returns: qvalues, pvalues
	"""

	LG.info("Association testing...")

        if string.upper(self.asso_method)=='ML':
            #calc residulas on Y
            Y_ = self._get_residuals(correct_trans=correct_trans)
            K  = None
        elif string.upper(self.asso_method)=='LMM':
            #use raw Y, LMM is taking care of this
            Y_= self.pheno
            [K,hyperparams_o] = self.get_K_testing(correct_trans=correct_trans,
                                                correct_confounders = True,
						noise=False)

	# find associations running a simple (and fast) linear model
	pval = testing.interface(self.snps, Y_,K, I = None,
				 model=string.upper(self.asso_method), parallel = self.parallelize,
				 file_directory = None, jobs = self.Njobs)[0]
 	# convert to qvalues
        qval = fdr.qvalues(pval)
        return qval,pval
    

    def getLatent(self,filter_active=False,ard_reweight=False,pca_rotation=False,rotate_varimax=False):
        """return active LVs
        ard_reweight: reweight factors by ARD relevances (False)
        pca_rotation: rotate to PCA solution (False)
        rotate_varimax: varimax rotation of factors (False)
        """

        X = self.hyperparams['x'].copy()
        I_ard = self.get_covar_param_index(self.factor_cf)

	aw_X_var = X.var(axis=0)
	aw_X_var /= aw_X_var.max()

	cutoff = self.cutoff_factors
	active = (aw_X_var > cutoff)

        # ARD reweighting and PCA rotation?
        if ard_reweight:
            I_ard = self.get_covar_param_index(self.factor_cf)
            a_w = 1/self.hyperparams['covar'][I_ard]
            X = SP.multiply(SP.sqrt(a_w), X)
	    
        if pca_rotation:
	    X -= X.mean(axis=0)
            X = GPLVM.PCA(SP.dot(X, X.T),self.n_factors)[0]

        if rotate_varimax:
            U, s, Vh = SP.linalg.svd(X, full_matrices = False)
            X = SP.dot(U, SP.diag(s))
	    
        # restrict to subset
        if filter_active:
            X = X[:,active]
            return X, active 

	return X


    def add_SNP_factor_associations(self):
        """
	Scans the SNP to find which ones overlap with the latent confounders.
	"""
	S = self.snps
	# get weighted latent variables
	X = self.getLatent(ard_reweight=True,pca_rotation=True)
        # Center the latent variables and the SNPs
	X -= X.mean(axis = 0)
        # Center SNPs
	S -= S.mean(axis = 0)

        # Get the covariance function that accounts for the current set of associations
        [K,hyperparams_o] = self.get_K_testing(correct_trans=True,correct_confounders=False,noise=False)

	# Association testing:
        PV = testing.interface(S, X, K, I = None, model = "LMM",parallel = self.parallelize, jobs = self.Njobs,
			       file_directory=self.testing_directory)[0]

        # Number of tests conducted
        Neff = self.n_factors*self.nS
        QV = fdr.qvalues(PV,m=Neff)

	# If needed, save diagnostic plots
        if self.verbose_plot_dir is not None:
            import pylab as PL
	    PL.figure(2)
	    PL.clf()
            PL.plot(self.candidate_associations, [1 for x in self.candidate_associations],  "rx")
            PL.savefig(os.path.join(self.verbose_plot_dir,'candidates_%d.png' % (self.train_iterations)))
            PL.figure(3)
	    PL.clf()
	    QVs = SP.array([SP.sort(QV[:,i]) for i in xrange(QV.shape[1])])
	    PL.plot(QVs.T)
            PL.savefig(os.path.join(self.verbose_plot_dir,'QVs_%d.png' % (self.train_iterations)))


	# Set the qvalue of the current associations to 1
	QV[self.candidate_associations,:] = 1

	# Greedily construct addition set by adding the BEST (lowest qv) SNP for each factor
	# (if significant)
	new_candidates = []
	for i in xrange(QV.shape[1]):
	    i_best = QV[:,i].argmin()
	    qv_best = QV[:,i].min()
	    # if significant, add it
	    if qv_best<=self.FDR_addition_associations:
		new_candidates.append(i_best)
		# and set the corrisponding qvalue to 1 
		QV[i_best,:] = 1

	# Add candidates
	Nc_old = len(self.candidate_associations)
	self.candidate_associations.extend(new_candidates)
	Nc = len(self.candidate_associations)

	dl = Nc-Nc_old
	#check everything is fine
	assert len(SP.unique(self.candidate_associations))==len(self.candidate_associations), 'added non-unique snps'
	LG.info("Candidates: %d, added: %d" % (Nc,dl))
	return dl

    
    def add_snp_factor_interactions(self):
        """
        Scan for interactions between all SNPs and genes, factors
        """

        #interactions_iter: conting number of effective iterations so far
	self.interactions_iter += 1

	correct_interactions = True
	
	# get the covariance matrix
	K = self.Kpop
	
        # get latent factors and rotate to PCA
	X = self.getLatent(ard_reweight = True, pca_rotation = True, rotate_varimax = False)

	inter = interaction_scan.get_candidate_interactions(self.pheno, self.snps, X, K,
							    interactions_iter = self.interactions_iter,
							    pv_max_stored = self.pv_max_stored,
							    topn_interactions = self.topn_interactions,
							    FDR_addition_interactions = self.FDR_addition_interactions,
							    candidate_interactions = self.candidate_interactions,
							    correct_interactions = correct_interactions,
							    parallel = self.parallelize,
							    Njobs = self.Njobs)

	Nint_new, self.candidate_interactions, int_pv_qv = inter
	self.interactions_pv, self.interactions_qv = int_pv_qv
	
	return Nint_new
    


    def train(self,Niter=1,Niter_interact=3,**kw_args):
        """
	Trains the latent variable model while adding associations and/or interactions

	Parameters:

	Niter = number of training iterations (if 1, just trains the GPLVM, without adding associations
	or interactions)

	"""
	
	LG.info("Training PANAMA with an association FDR of %.2f" % self.FDR_addition_associations)

	if self.model == "gplvm_interact":
	    LG.info("LIMMI's interaction FDR is set to %.2f, adding at most %d interactions" % (self.FDR_addition_interactions,
												self.topn_interactions))
	
        # retrain once
	retrain = True
	# iteration counter
	it = 0

	# init the mode variable
	mode = None

        # current mode of covariance additions
        if self.add_associations:
	    LG.info("add_associations mode active")
            mode = 'add_associations'
        elif self.add_interactions:
	    LG.info("add_interactions mode active")	    
            mode = 'add_interactions'
        # ok, now the actual loop
	while (retrain and it < Niter):
            # association scan and interactions require one training iteration
            if self.train_iterations > 0:
		# adds association terms
                if mode == 'add_associations':
                    LG.info("adding factors-SNPs associations")
		    asso_added = self.add_SNP_factor_associations()
		    if (asso_added==0):
                        # switch mode to interactions?
                        if self.add_interactions:
			    LG.info("All associations added, switching to adding interactions")
                            mode = 'add_interactions'
                            Niter = it+Niter_interact
                        else:
                            break
                elif mode == 'add_interactions':
                    LG.info("Adding factors-SNPs interactions")
                    int_added = self.add_snp_factor_interactions()

                    if int_added == 0:
                        break
			
            LG.info('Train iteration %d/%d' % (it+1, Niter))
	    # If we are at the first iteration, find an estimate for K
	    if self.train_iterations == 0:
		if not self.fixed_latent_dim:
		    # determine latent dimensionality in first iteration
		    self._init_gp()
		    K = self.findK()
		    self.n_factors = K
		    # kill hyperparmas again because number of factors changed!
		    self.hyperparams = None
		    LG.info("PCA-initialized dimensionality: %d" % K)
		else:
		    LG.info("User-supplied latent dimensionality: %d" % self.n_factors)

            #note: we swtiched iterative factors on, which trians factors and hyperparameters in turns:
            self.train_gplvm(iterative_factors=False,**kw_args)
            # increase iteration counter
            self.train_iterations +=1
	    it+=1
            # store lml
            self.lmls.append(self.lMl())
        pass
        
                          
    def train_gplvm(self, gradcheck = False, Nrestarts = 1, diagnostics = True,iterative_factors=False,n_iterative_factors=3, **kw_args):
        """
	Trains the GPLVM model without priors.
	Nrestarts specifies the number of random restarts
	iterative_factors: train factors and hyperparameters in turns
	"""
        def opt_hyper(iterative_factors=False):
            if iterative_factors:
                #note: we always refit the likelihood as there seems to be no good reason to not do so:
                #filter for learning covar params:
                Ifilter_covar = {'covar': SP.ones(len(self.hyperparams['covar']),dtype='bool'),
                                 'lik': SP.ones(len(self.hyperparams['lik']),dtype='bool'),
                                 'x': SP.zeros(self.hyperparams['x'].shape,dtype='bool')}

                #filter for learning factors
                Ifilter_factors = {'covar': SP.zeros(len(self.hyperparams['covar']),dtype='bool'),
                                 'lik': SP.ones(len(self.hyperparams['lik']),dtype='bool'),
                                 'x': SP.ones(self.hyperparams['x'].shape,dtype='bool')}
                #iterate
                opt_params = self.hyperparams
                for i in xrange(n_iterative_factors):
                    [opt_params,opt_lml]= GPLVM.opt_hyper(self.gp,opt_params,gradcheck=gradcheck,bounds=self.hyperparams_bounds,priors = None,Ifilter=Ifilter_covar, **kw_args)
		    # opt_params['x'] = self.getLatent(ard_reweight=True, rotate_varimax = True)
                    [opt_params,opt_lml]= GPLVM.opt_hyper(self.gp,opt_params,gradcheck=gradcheck,bounds=self.hyperparams_bounds,priors = None,Ifilter=Ifilter_factors, **kw_args)
                #end for
                
            else:
                [opt_params,opt_lml]= GPLVM.opt_hyper(self.gp,self.hyperparams,gradcheck=gradcheck,bounds=self.hyperparams_bounds,priors = None, **kw_args)

            return  [opt_params,opt_lml]

        
        
        lmls = []
        params = []
	i = 0 
	while i < Nrestarts:
            # trains the GPLVM without priors
	    self._init_gp()
	    i += 1
            #opt
            [opt_params,opt_lml] = opt_hyper(iterative_factors=iterative_factors)
            lmls.append(opt_lml)
            params.append(opt_params)

        # save restart results
        self.restart_results = {'lmls': SP.array(lmls), 'params': SP.array(params)}

	# find best run and use the parameters from that
        ibest = self.restart_results['lmls'].argmin()
        opt_model_params = self.restart_results['params'][ibest]
        self.hyperparams = opt_model_params

        # invalidate caches
        self.QV_asso = None

	# If a plot directory is supplied, plots the factor activations
        if self.verbose_plot_dir is not None:
	    import pylab as PL
            PL.figure(4)
            PL.clf()
            I_factor = self.get_covar_param_index(self.factor_cf)
            weights = 1/self.hyperparams['covar'][I_factor]
            x       = self.hyperparams['x']
            weightseff   = weights*x.std(axis=0)
            PL.bar(SP.arange(len(weightseff)),weightseff)
            PL.savefig(os.path.join(self.verbose_plot_dir,'active_factors%d.pdf' % (self.train_iterations)))


    #####PRIVATE METHODS######
    def _init_latent(self):
        """
	Returns starting point for latent variables
	"""
        # initialise latent variables
        if self.gplvm_init=='PCA':
            [Spca,Wpca] = GPLVM.PCA(self.pheno,self.n_factors)
            self.X0 = Spca.copy()
            x_latent = Spca
        elif self.gplvm_init=='RANDN':
            x_latent = SP.random.randn(self.nN,self.n_factors) * 0.1
        elif self.gplvm_init=='PCA_RANDN':
            # Noisy PCA
            [Spca,Wpca] = GPLVM.PCA(self.pheno,self.n_factors)
            self.X0 = Spca.copy()
            x_latent = Spca
            x_latent += 0.1*SP.random.randn(self.nN,self.n_factors)
        elif isinstance(self.gplvm_init,SP.ndarray):
            assert self.gplvm_init.shape==(self.nN,self.n_factors)
            x_latent = self.gplvm_init.copy()
        else:
            raise Excetion('unknown initialization method')
        return x_latent

        
    def _init_gp(self, hyperparams_latent = None):
        """
	Initialization of gp models and covariances. If xlantent and/or hyperparams_latent
	are supplied, it skips the initialization with PCA/randn and uses the provided
	latent variables and hyperparameters
	"""

        # all snps
        self.snp_dimensions   = SP.arange(0,self.nS)
        # then covariates
        self.covar_dimensions = SP.arange(self.nS,self.nS+self.nC)
        # then factors
        self.factor_dimensions = SP.arange(self.nS+self.nC,self.nS+self.nC+self.n_factors)

        snp_dimensions = self.snp_dimensions
        factor_dimensions = self.factor_dimensions

      
        # 2. construct covariances                
        # construct GPLVM model
        covar = []

        # 2.0 mean CF
        self.mean_cf  = mu.MuCF(self.nN)
        self.mean_cf_covar = SP.array([1E-3])
        self.mean_cf_covar_bound = [(-SP.inf,10)]
        self.mean_cf_covar_prior = [[lnpriors.lnuniformpdf,[]]]

        # 2.1 noise init	
	if not self.FA_noise:
	    self.noise_cf_covar = SP.array([self.sigma])
	    self.noise_cf_covar_bound = [(-SP.inf,10)]
	    self.noise_cf_covar_prior = [[lnpriors.lnuniformpdf,[]]]
        else:
	    self.noise_cf_covar = SP.log(SP.ones(self.nD)+self.sigma*SP.random.randn(self.nD))
	    self.noise_cf_covar_bound = [(-SP.inf,10)] * self.nD
	    self.noise_cf_covar_prior = [[lnpriors.lnuniformpdf,[]]] * self.nD

        # 2.2 global genetics CF (population only)
        if self.use_pop_cf:
	    assert self.Kpop != None, "use_pop_cf is true but no population structure covariance was supplied"	    
            self.pop_cf = fixed.FixedCF(K=self.Kpop)
            self.pop_cf_covar = SP.array([1])
            self.pop_cf_covar_bound = [(-SP.inf,10)]
            self.pop_cf_covar_prior = [[lnpriors.lnuniformpdf,[]]]

        # 2.3 Covariates CF
        self.cov_cf = linear.LinearCFARD(dimension_indices = self.covar_dimensions)
        self.cov_cf_covar = SP.array([SP.exp(1)]*self.nC)
        self.cov_cf_covar_bound = [(1E-3,SP.inf)]*self.nC
        self.cov_cf_covar_prior = [[lnpriors.lnuniformpdf,[]]] * self.nC
       
        # 2.3 SNP covariance function
        self.snp_cf = linear.LinearCF(dimension_indices = snp_dimensions[self.candidate_associations])
        self.snp_cf_covar = SP.array([1]*len(self.candidate_associations))
        self.snp_cf_covar_bound = [(-SP.inf,10)]*len(self.candidate_associations)
        self.snp_cf_covar_prior = [[lnpriors.lnuniformpdf,[]]]*len(self.candidate_associations)
               
        # 2.4 factor covariance function
        if self.ard:
            # use new ARD covariance function, this stilll operatores inlog space!
            # weights in new covariances are just 1/logtheta[i], so no exp!
            self.factor_cf = linear.LinearCFARD(dimension_indices = factor_dimensions)
            # use SP.exp such that log from this is 1 again:
	    if hyperparams_latent == None:
		self.factor_cf_covar = SP.array([SP.exp(100)]*self.n_factors)
	    else:
		self.factor_cf_covar = SP.exp(hyperparams_latent)
            self.factor_cf_covar_bound = [(1E-3,SP.inf)]*self.n_factors
            self.factor_cf_covar_prior = [[lnpriors.lnL1,[self.lam]]]*self.n_factors
            pass
        else:
            self.factor_cf = linear.LinearCFISO(dimension_indices = factor_dimensions)

	    if hyperparams_latent == None:
		self.factor_cf_covar = SP.array([1])
	    else:
		self.factor_cf_covar = hyperparams_latent
		
            self.factor_cf_covar_bound = [(-SP.inf,10)]
	    self.factor_cf_covar_prior = [[lnpriors.lnL1,[self.lam]]]


        self.covar_list = []
        self.covar_cf_list = []
        self.covar_cf_bounds = []
        self.covar_cf_priors = []
        
        #optinal: mean_cf and population CF:
        if self.use_mean_cf:
            self.covar_list.extend([self.mean_cf])
            self.covar_cf_list.extend([self.mean_cf_covar])
            self.covar_cf_bounds.extend(self.mean_cf_covar_bound)
            self.covar_cf_priors.extend(self.mean_cf_covar_prior)
        if self.use_pop_cf:
            self.covar_list.extend([self.pop_cf])
            self.covar_cf_list.extend([self.pop_cf_covar])
            self.covar_cf_bounds.extend(self.pop_cf_covar_bound)
            self.covar_cf_priors.extend(self.pop_cf_covar_prior)

        #add covariance CF if needed
        if self.nC>0:
            self.covar_list.extend([self.cov_cf])
            self.covar_cf_list.extend([self.cov_cf_covar])
            self.covar_cf_bounds.extend(self.cov_cf_covar_bound)
            self.covar_cf_priors.extend(self.cov_cf_covar_prior)
            

        #add snp cf and noise CF
        self.covar_list.extend([self.snp_cf])
        self.covar_cf_list.extend([self.snp_cf_covar])
        self.covar_cf_bounds.extend(self.snp_cf_covar_bound)
        self.covar_cf_priors.extend(self.snp_cf_covar_prior)
	

	# Factor CF common to PANAMA and LIMMI
        if self.model == 'gplvm' or self.model == "gplvm_interact":
            self.covar_list.extend([self.factor_cf])
            self.covar_cf_list.extend([self.factor_cf_covar])
            self.covar_cf_bounds.extend(self.factor_cf_covar_bound)
            self.covar_cf_priors.extend(self.factor_cf_covar_prior)

	    # specific to LIMMI
	    if self.model == "gplvm_interact":
		self.interact_cf = LinearInteractCF(dimension_indices = factor_dimensions,
						    interact_dimension_indices = snp_dimensions,
						    interact_pairs = copy.deepcopy(self.candidate_interactions))
		self.interact_cf_covar = SP.array([1]*len(self.candidate_interactions))
		# TODO: is this right???
		self.interact_cf_covar_bound = [(-SP.inf,10)]*len(self.candidate_interactions)
		self.interact_cf_covar_prior = [[lnpriors.lnuniformpdf,[]]]*len(self.candidate_interactions)

		self.covar_list.extend([self.interact_cf])
		self.covar_cf_list.extend([self.interact_cf_covar])
		self.covar_cf_bounds.extend(self.interact_cf_covar_bound)
		self.covar_cf_priors.extend(self.interact_cf_covar_prior)
                # print "switching on add_interactions because of gplvm_interact"
                # self.add_interactions = True

	       		
        else:
            raise Exception("unknown covariance model")

        # create sum CF
        self.covariance_gplvm = combinators.SumCF(tuple(self.covar_list))
	    
	# construct hyperparameters
        log_covar = SP.log(SP.concatenate(self.covar_cf_list))

        # Initialize observed inputs
        x_obs = self.snps
        x_covar = self.covariates
	# Initialize latent variables
	x_latent = self._init_latent()

	self.hyperparams_init = {'covar':log_covar,'x':x_latent, 'lik': self.noise_cf_covar}
        if (self.hyperparams is None) or (not self.keep_latent):
            self.hyperparams = copy.deepcopy(self.hyperparams_init)
        else:
            self.hyperparams['covar'] = self.hyperparams_init['covar']
	    self.hyperparams['lik'] = self.hyperparams_init['lik']
	
	
        # create GP object
	if not self.FA_noise:
	    self.gp = GPLVM.GPLVM(covar_func = self.covariance_gplvm, likelihood = lik.GaussLikISO())
	else:
	    self.gp = GPLVM_ARD.GPLVMARD(covar_func = self.covariance_gplvm,
					 likelihood = lik.GaussLikARD(n_dimensions = self.nD))


        # set x_latent in this state to 0.
        # this ensures that the first iteration latent variables (before training) are not used
        x_latent0 = SP.zeros_like(x_latent)
             
        self.X = SP.concatenate((x_obs,x_covar,x_latent0),axis=1)
        # set active dimensions in covariance function:
        self.gp.setData(x=self.X,y=self.pheno,gplvm_dimensions=factor_dimensions)
        # create bounds
        self.hyperparams_bounds = {'covar': SP.array(self.covar_cf_bounds),
				   'lik': SP.array(self.noise_cf_covar_bound)}
        # create prior
        self.hyperparams_priors = {'covar': SP.array(self.covar_cf_priors),
				   'lik': SP.array(self.noise_cf_covar_prior)}

	# self.get_explained_var()
        pass



import sys, os, string, pickle, copy, pdb
sys.path.append('./../../pygp')
sys.path.append('./../../')
#gplvm model
from pygp.covar import linear, noise, combinators, fixed, mu
import pygp.gp.gplvm as GPLVM
import pygp.gp.gplvm_ard as GPLVM_ARD
# customized covariance function:
from interaction_covar import *
import pygp.priors.lnpriors as lnpriors
import pygp.likelihood as lik
import linear_regression as linreg
import logging as LG
import scipy as SP
import numpy as N
from panama.utilities import fdr
import scipy.stats as st
import scipy.sparse as sparse
import testing


def scan(Y, S, X, K, pv_min, all_X, sparse_storage, interactions_iter = None, parallel = False, Njobs = 0,
	 correct_interactions = None, **kw_args):
    """sub function doing the actual scan"""

    interactions_qv, interactions_pv = None, []
    
    if not all_X:
	PV = []
	INDEX = []

	# sometimes (like when we are simulating) we don't want to store sparse QVs, so
	# let's just compute the whole thing
	if not sparse_storage:
	    PV = SP.empty((S.shape[1], Y.shape[1], X.shape[1]))
	    LODS = SP.empty((S.shape[1], Y.shape[1], X.shape[1]))

	# only for testing
	# Xt = pickle.load(open('/mnt/nicolo_storage/limmi/simulation/results/true_LVs.pickle', 'r'))
	# K = SP.dot(Xt, Xt.T)

	if K == None:
	    K = SP.zeros((Y.shape[0],Y.shape[0]))
	    
	for i in range(X.shape[1]):

	    [lod,pv_lmm] = testing.interface(S, Y, K, I = X[:,i:i+1], model = "LMM",
				       parallel = parallel,
				       jobs = Njobs, return_fields= ['lod','pv'], Ftest = True, **kw_args)

	    if sparse_storage:
		# 1. figure out which pv are < min
		Irel = pv_lmm < pv_min
		# add random number of PVs for histogram calculation etc.
		# fix: add 100,000 random PV observations
		fraction = 100000.0/Irel.size
		Irandom = SP.random.rand(Irel.shape[0],Irel.shape[1])<fraction

		# store pvalues for histograms and stuff
		interactions_pv.extend(pv_lmm[Irandom])

		pv_rel = pv_lmm[Irel]
		index = SP.nonzero(Irel)
		index_snps = index[0]
		index_genes  = index[1]
		index_factors = SP.ones_like(index_genes)*i
		index_full = SP.concatenate((index_snps[:,SP.newaxis],index_genes[:,SP.newaxis],index_factors[:,SP.newaxis]),axis=1)
		INDEX.extend(index_full)
		PV.extend(pv_rel)
	    else:
		PV[:,:,i] = pv_lmm
		LODS[:,:,i] = lod		    

	# calc effective number of tests
	n_effective_tests = X.shape[1] * Y.shape[1] * S.shape[1] * interactions_iter

	if sparse_storage:
	    #convert everything into arrays
	    PV = SP.array(PV)
	    INDEX = SP.array(INDEX)

	    QV = fdr.qvalues(PV, m = n_effective_tests, fix_lambda = pv_min)
	    QVMi = SP.concatenate((SP.array(INDEX), SP.array(QV)[:, SP.newaxis]), axis=1)

	    # use more sane types (it used to be all floats64), this 
	    dtype = N.dtype({'names':['SNP','gene', 'factor', 'qv'],
			     'formats':[N.uint,N.uint,N.uint,N.float64]})
	    QVMi = N.rec.fromarrays(QVMi.swapaxes(1,0), dtype=dtype)

	    # sort by QV
	    Is = QV.argsort()
	    QV = QV[Is]
	    INDEX = INDEX[Is]

	else:
	    QV = fdr.qvalues(PV, m = n_effective_tests)
	    QVMi = QV.min(axis=2) # minimize across factors
	    # QVMi = QV.min(axis=1) # minimize across genes


	# store qvalues
	interactions_qv = QVMi

    #return Index: snps, genes, factors
    inter_results = (interactions_pv,
		     interactions_qv) 
    
    return [INDEX,QV,inter_results]


def get_candidate_interactions(Y, S, X, K, pv_max_stored = 1.0, topn_interactions = 3, FDR_addition_interactions = 0.01, candidate_interactions = [], **kw_args):    

    #find interactions
    fit_all_X = False

    # MLM based scan
    # see details in function on meaning of index
    sparse_storage = True

    if pv_max_stored == None:
	pv_min = FDR_addition_interactions
    else:
	pv_min = pv_max_stored

    [INDEX,QV,inter_results] = scan(Y, S, X, K, pv_min, fit_all_X, sparse_storage, **kw_args)

    # QV are sorted by size already
    
    if sparse_storage:
	# starting index on those guys that are still in the game
	Iok = SP.ones(QV.shape[0],dtype='bool')

	print "Significant QVs", (QV < FDR_addition_interactions).sum()

	# find new candidate guys to append:
	new_candidate_interactions = []
	# iteratively cycle throug best SNP and factor
	while True:

	    # only add the top n interactions
	    if len(new_candidate_interactions) == topn_interactions:
		break

	    # check still candidates in the game
	    if not Iok.any():
		break

	    # take the current best guy (they are ordered)
	    ic = SP.nonzero(Iok)[0][0]

	    qvm = QV[ic]
	    if qvm > FDR_addition_interactions:
		#break if best guy worse than cutoff
		break

	    # get factor/snp index
	    [i_snp,i_gene,i_factor] = INDEX[ic]

	    # append to candidates
	    new_candidate_interactions.append((i_factor, i_snp))

	    # adjust IOK to skip that factor and SNP in followup rounds
	    i_bad_snp = (INDEX[:,0] == i_snp)
	    i_bad_factor = (INDEX[:,2] == i_factor)

	    # Iok[i_bad_snp | i_bad_factor] = False
	    Iok[i_bad_snp] = False

	    print "Remaining sign. QVs: ", (QV[Iok] < FDR_addition_interactions).sum()
    else:
	# set QV for SNPS that are already in candidates to 1
	# if len(candidate_interactions) != 0:
	#    interacting_SNPs = SP.array(candidate_interactions)[:, 1]
	#    QV[interacting_SNPs,:] = 1

	QV = QV.min(axis=1) # minimize across genes
	new_candidate_interactions = []
	#iteratively cycle throug best SNP and factor

	while True:
	    # only add the top n interactions
	    if len(new_candidate_interactions) == topn_interactions:
		break

	    # find best factor/snp combination
	    qvm = QV.min()
	    qvmw = SP.where(QV==qvm)
	    
	    if qvm > FDR_addition_interactions:
		#break if no one left
		break

	    # factor/snp index
	    [i_snp,i_factor] = [qvmw[0][0],qvmw[1][0]]


	    # this could be useful in the case of LD (blanks region)
	    # low_snp = max(0, i_snp-10)
	    # hi_snp = min(i_snp+10, self.snps.shape[1])	    
	    # QV[low_snp:hi_snp,:] = 1.

	    # remove from QV table
	    QV[i_snp,:] = 1.
	    QV[:,i_factor] = 1.

	    # append to candidates
	    new_candidate_interactions.append((i_factor, i_snp))

    Nint_new = len(new_candidate_interactions)
    Nint_old = len(candidate_interactions)
    print new_candidate_interactions
    candidate_interactions.extend(new_candidate_interactions)
    LG.info("Interaction candidates: %d, added %d" % (Nint_old, Nint_new))
    
    return Nint_new, candidate_interactions, inter_results




class Tango(object):
    def __init__(self):
	self.Aluminium6="#2e3436"
	self.Aluminium5="#555753"
	self.Aluminium4="#888a85"
	self.Aluminium3="#babdb6"
	self.Aluminium2="#d3d7cf"
	self.Aluminium1="#eeeeec"

	self.lightPurple="#ad7fa8"
	self.mediumPurple="#75507b"
	self.darkPurple="#5c3566"

	self.lightBlue="#729fcf"
	self.mediumBlue="#3465a4"
	self.darkBlue= "#204a87"

	self.lightGreen="#8ae234"
	self.mediumGreen="#73d216"
	self.darkGreen="#4e9a06"

	self.lightChocolate="#e9b96e"
	self.mediumChocolate="#c17d11"
	self.darkChocolate="#8f5902"

	self.lightRed="#ef2929"
	self.mediumRed="#cc0000"
	self.darkRed="#a40000"

	self.lightOrange="#fcaf3e"
	self.mediumOrange="#f57900"
	self.darkOrange="#ce5c00"

	self.lightButter="#fce94f"
	self.mediumButter="#edd400"
	self.darkButter="#c4a000"

	# self.darkList = [self.darkBlue, self.darkRed, self.darkGreen, self.darkOrange,
	# 		 self.darkButter, self.darkPurple, self.darkChocolate, self.Aluminium6]
	self.darkList = [self.darkBlue, self.darkRed, self.darkGreen, self.darkOrange,
			 self.Aluminium5, self.darkPurple, self.darkChocolate, self.Aluminium6]
	self.mediumList = [self.mediumBlue, self.mediumRed, self.mediumGreen, self.mediumOrange,
			   self.mediumButter, self.mediumPurple, self.mediumChocolate, self.Aluminium5]
	self.lightList = [self.lightBlue, self.lightRed, self.lightGreen, self.lightOrange, self.lightButter,
			  self.lightPurple, self.lightChocolate, self.Aluminium4]


    def nextDark(self):
	self.darkList.append(self.darkList.pop(0))
	return self.darkList[-1]
    def nextMedium(self):
	self.mediumList.append(self.mediumList.pop(0))
	return self.mediumList[-1]
    def nextLight(self):
	self.lightList.append(self.lightList.pop(0))
	return self.lightList[-1]



##ROC
#helper class for the evaluation of ROC curves and area under ROC

import scipy as S
import numpy as N
import pdb


def roc(labels, predictions):
    """roc - calculate receiver operator curve
    labels: true labels (>0 : True, else False)
    predictions: the ranking generated from whatever predictor is used"""
    #1. convert to arrays
    labels = S.array(labels).reshape([-1])
    predictions = S.array(predictions).reshape([-1])

    #threshold
    t = labels>0
    
    #sort predictions in desceninding order
    #get order implied by predictor (descending)
    Ix = S.argsort(predictions)[::-1]
    #reorder truth
    t = t[Ix]

    #compute true positiive and false positive rates
    tp = S.double(N.cumsum(t))/t.sum()
    fp = S.double(N.cumsum(~t))/(~t).sum()

    #add end points
    tp = S.concatenate(([0],tp,[1]))
    fp = S.concatenate(([0],fp,[1]))

    return [tp,fp]

def auroc(labels=None,predictions=None,tp=None,fp=None,partial=False):
    """auroc - calculate area under the curve from a given fp/rp plot"""

    if labels is not None:
        [tp,fp] = roc(labels,predictions)

    if partial != False:
	n = N.abs(fp - partial).argmin()
	tp = tp[:n]
	fp = fp[:n]	

    n = tp.size	
    auc = 0.5*((fp[1:n]-fp[0:n-1]) * (tp[1:n]+tp[0:n-1])).sum()

    if partial != False:
	return auc/partial
    else:
	return auc



import scipy.stats as st
import numpy as NP
cimport numpy as NP
from libcpp cimport bool

cdef extern from "numpy/arrayobject.h":
    ctypedef int intp
    ctypedef extern class numpy.ndarray [object PyArrayObject]:
        cdef char *data
        cdef int nd
        cdef intp *dimensions
        cdef intp *strides
        cdef int flags


###nLLeval
cdef extern from "lmm.h": 
    double c_nLLeval "nLLeval" (double ldelta,double* dUY,double* dUX,double* dS,int n,int d)
    double c_optdelta "optdelta" (double* dUY,double* dUX,double* dS,int n,int d,int numintervals,double ldeltamin,double ldeltamax)
    void c_train_associations "train_associations" (double* dLOD,double* dX,double* dY,double* dK,double* dC,int nn,int ns,int np,int nc,int numintervalsAlt,double ldeltaminAlt,double ldeltamaxAlt,int numintervals0,double ldeltamin0,double ldeltamax0)
    void c_train_interactions "train_interactions" (double* dPVALS,double* dLOD,double* dX,double* dY,double* dK,double* dC,double* dI,int nn,int ns,int np,int nc,int ni,int numintervalsAlt,double ldeltaminAlt,double ldeltamaxAlt,int numintervals0,double ldeltamin0,double ldeltamax0,bool refit_delta0_snp, bool use_ftest)

    void c_train_associations_SingleSNP "train_associations_SingleSNP" ( double* PV, double* LL, double* ldelta, double* X, double* Y, double* U, double* S, double* C, int nn,int ns,int np,int nc, int numintervals, double ldeltamin, double ldeltamax)

def ndarrayF64toC(ndarray A):
    #transform numpy object to enforce Fortran contiguous byte order 
    #(meaining column-first order for Cpp interfacing)
    return NP.asarray(A, order="F")
#   OLD and c-tingous version
#     assert (<object>A).dtype == NP.float64
#     if not (<object>A).flags["C_CONTIGUOUS"]:
#         Ac = A.copy('C')
#     else:
#         Ac = A
#     return Ac
        
def train_associations(ndarray X,ndarray Y,ndarray K,ndarray C,int numintervalsAlt=0,double ldeltaminAlt=-1,double ldeltamaxAlt=1,int numintervals0=100,double ldeltamin0=-5,double ldeltamax0=5,calc_pval=True):
    #1. get dimensions
    cdef int nn = X.dimensions[0]
    cdef int ns = X.dimensions[1]
    cdef int np = Y.dimensions[1]
    cdef int nc = C.dimensions[1]
    #2. create result structure for LOD
    cdef NP.ndarray LOD=NP.zeros([np,ns],dtype=NP.float)
    #3. get raw data pointers
    X = ndarrayF64toC(X)
    Y = ndarrayF64toC(Y)
    K = ndarrayF64toC(K)
    C = ndarrayF64toC(C)
    LOD = ndarrayF64toC(LOD)  
    cdef double *dY = <double* > Y.data
    cdef double *dX = <double* > X.data
    cdef double *dK = <double* > K.data
    cdef double *dC = <double* > C.data
    cdef double *dLOD = <double*> LOD.data               
    #4. call C
    c_train_associations(dLOD,dX,dY,dK,dC,nn,ns,np,nc,numintervalsAlt,ldeltaminAlt,ldeltamaxAlt,numintervals0,ldeltamin0,ldeltamax0)
    #5. return
    if calc_pval:
        pval = st.chi2.sf(2*(LOD),1)
        return LOD, pval
    else:
        return LOD
   
def train_interactions(ndarray X,ndarray Y,ndarray K,ndarray C,ndarray Inter,int numintervalsAlt=0,double ldeltaminAlt=-1,double ldeltamaxAlt=1,int numintervals0=100,double ldeltamin0=-5,double ldeltamax0=5,bool refit_delta0_snp=False,use_ftest=False):
    #1. get dimensions
    cdef int nn = X.dimensions[0]
    cdef int ns = X.dimensions[1]
    cdef int np = Y.dimensions[1]
    cdef int nc = C.dimensions[1]
    cdef int ni = Inter.dimensions[1]
    #2. create result structure for final pvalues
    cdef NP.ndarray PV=NP.zeros([np,ns],dtype=NP.float,order="F")
    cdef NP.ndarray LOD=NP.zeros([np,ns],dtype=NP.float,order="F")	
    #3. get raw data pointers
    X = ndarrayF64toC(X)
    Y = ndarrayF64toC(Y)
    K = ndarrayF64toC(K)
    C = ndarrayF64toC(C)
    Inter = ndarrayF64toC(Inter)
    PV = ndarrayF64toC(PV)  
    LOD = ndarrayF64toC(LOD)	
    cdef double *dY = <double* > Y.data
    cdef double *dX = <double* > X.data
    cdef double *dK = <double* > K.data
    cdef double *dC = <double* > C.data
    cdef double *dInter = <double*> Inter.data
    cdef double *dPV = <double*> PV.data               
    cdef double *dLOD = <double*> LOD.data               
    #4. call C
    c_train_interactions(dPV,dLOD,dX,dY,dK,dC,dInter,nn,ns,np,nc,ni,numintervalsAlt,ldeltaminAlt,ldeltamaxAlt,numintervals0,ldeltamin0,ldeltamax0,refit_delta0_snp,use_ftest)
    return LOD,PV

def train_associations_SingleSNP( ndarray X, ndarray Y, ndarray U, ndarray S, ndarray C, int numintervals = 100, double ldeltamin = -10.0, double ldeltamax = 10.0):
    #1. get dimensions
    cdef int nn = X.dimensions[0]
    cdef int ns = X.dimensions[1]
    cdef int np = Y.dimensions[1]
    cdef int nc = C.dimensions[1]
    #2. create result structure for final pvalues
    cdef NP.ndarray PV=NP.zeros([np,ns],dtype=NP.float,order="F")
    cdef NP.ndarray LL=NP.zeros([np,ns],dtype=NP.float,order="F")
    cdef NP.ndarray ldelta=NP.zeros([np,ns],dtype=NP.float,order="F")
    #3. get raw data pointers
    X = ndarrayF64toC(X)
    Y = ndarrayF64toC(Y)
    S = ndarrayF64toC(S)
    U = ndarrayF64toC(U)
    C = ndarrayF64toC(C)
    U = ndarrayF64toC(U)
    S = ndarrayF64toC(S)	
    PV = ndarrayF64toC(PV)  
    LL = ndarrayF64toC(LL)  
    ldelta = ndarrayF64toC(ldelta)  
    cdef double *dY = <double* > Y.data
    cdef double *dX = <double* > X.data
    cdef double *dU = <double* > U.data
    cdef double *dS = <double* > S.data
    cdef double *dC = <double* > C.data
    cdef double *dPV = <double*> PV.data
    cdef double *dLL = <double*> LL.data
    cdef double *dldelta = <double*> ldelta.data               
    #4. call C
    c_train_associations_SingleSNP( dPV, dLL, dldelta, dX, dY, dU, dS, dC, nn, ns, np, nc, numintervals, ldeltamin, ldeltamax)
    return PV,LL,ldelta

          
def optdelta(ndarray UY,ndarray UX,ndarray S):
    UX = ndarrayF64toC(UX)
    UY = ndarrayF64toC(UY)
    S = ndarrayF64toC(S)
    cdef double *dUX = <double*> UX.data
    cdef double *dUY = <double*> UY.data
    cdef double *dS = <double*> S.data
    cdef int n = UX.dimensions[0]
    cdef int d = UX.dimensions[1]
    return c_optdelta(dUY,dUX,dS,n,d,100,-10.,10.)


def nLLeval(double ldelta,ndarray UY,ndarray UX,ndarray S):
    UX = ndarrayF64toC(UX)
    UY = ndarrayF64toC(UY)
    S = ndarrayF64toC(S)
    cdef double *dUX = <double*> UX.data
    cdef double *dUY = <double*> UY.data
    cdef double *dS = <double*> S.data
    cdef int n = UX.dimensions[0]
    cdef int d = UX.dimensions[1]
    return c_nLLeval(ldelta,dUY,dUX,dS,n,d)
###





import sys, os

def get_num_threads():
    return os.getenv("OMP_NUM_THREADS")

def set_num_threads(num):
    """
    Sets the number of threads used by MKL. It is possible to control
    that using either MKL_NUM_THREADS or OMP_NUM_THREADS
    """
    
    os.environ["OMP_NUM_THREADS"] = str(num)

    return os.getenv("OMP_NUM_THREADS")

def set_dynamic_load(cond):
    """
    MKL_DYNAMIC being TRUE means that Intel MKL will always try to pick
    what it considers the best number of threads, up to the maximum
    specified by the user. MKL_DYNAMIC being FALSE means that Intel MKL
    will not deviate from the number of threads the user requested, unless
    there are reasons why it has no choice. The value of MKL_DYNAMIC is by
    default set to TRUE, regardless of OMP_DYNAMIC, whose default value
    may be FALSE.

    In general, you should set MKL_DYNAMIC to FALSE only under
    circumstances that Intel MKL is unable to detect, for example, when
    nested parallelism is desired where the library is called already from
    a parallel section. Please refer to "MKL_DYNAMIC" in the Intel MKL
    User's Guide for details        
    """

    if cond:
	os.environ['OMP_DYNAMIC'] = "TRUE"
    else:
	os.environ['OMP_DYNAMIC'] = "FALSE"

    return cond

def get_dynamic_load(cond):
    return os.getenv("OMP_DYNAMIC")

def get_num_cores():
    """
    Returns the number of cores available on the system
    """
    
    return os.sysconf("SC_NPROCESSORS_ONLN")



import scipy as SP
import pdb





def eigenstratImputation(X, imissX=None, maxval=2.0):
    """
    impute SNPs as in
    'Principle components analysis corrects for stratification in genome-wide association studies'
    Price et al 2006 (Nature Genetics) page 908, Section 'Methhods', 'Inference of axes of variation'

    Input:
      X:        [n_indiv * n_SNP] matrix of SNPs, missing values are SP.nan
      maxval:   maximum value in X (default is 2 assuming 012 encoding)

    Output:
      X_ret:        [n_ind * n_SNP] matrix of imputed and standardized SNPs
    """
    if imissX is None:
        imissX = SP.isnan(X)
    
    n_i,n_s=X.shape
    if imissX is None:
        n_obs_SNP=SP.ones(X.shape)
    else:    
        i_nonan=(~imissX)
        n_obs_SNP=i_nonan.sum(0)
        X[imissX]=0.0
    snp_sum=(X).sum(0)
    one_over_sqrt_pi=(1.0+snp_sum)/(2.0+maxval*n_obs_SNP)
    one_over_sqrt_pi=1./SP.sqrt(one_over_sqrt_pi*(1.-one_over_sqrt_pi))
    snp_mean=(snp_sum*1.0)/(n_obs_SNP)
    X_ret=X-snp_mean
    X_ret*=one_over_sqrt_pi
    if imissX is not None:
        X_ret[imissX]=0.0
    return X_ret





import lmm
#import lmm
import sys
import scipy.linalg as LA
import cPickle
#import standard fast python lmm
sys.path.append('./../core')
import lmm_fast
import scipy as SP
import pdb
#import pylab as PL

#TODO
#load data with SNPS x and phenotype matrix Y
X, Y, K = cPickle.load(open("S_Y_K_testing.pickle", "r"))
#run standard lmm
#X = SP.random.randn(X.shape[0],10000)

n,s=X.shape;
n_pheno=Y.shape[1];
S,U=LA.eigh(K);
UY=SP.dot(U.T,Y);
UX=SP.dot(U.T,X);


ldelta = +8

import time

#need to create copies to ensure it is all properly lined up in memory
#TODO: I am sure this issues is common to cython stuff and other people have similar problems
_UX = SP.array(UX[:,0:2])
_UY = SP.array(UY[:,0])

if 1:
    print "testing delta opt"
    delta0 = lmm_fast.optdelta(_UY,_UX,S)
    delta1 = lmm.optdelta(_UY,_UX,S)
    print "%.2f versus %.2f" % (delta0,delta1)

if 1:
    print "testing eval on all SNPs"
    for i in xrange(UX.shape[1]):
        _UX = UX[:,i:i+1]
        _UY = SP.array(UY[:,0])
        lml0=lmm_fast.nLLeval(ldelta,_UY,_UX,S)
        lml1=lmm.nLLeval(ldelta,_UY,SP.array(_UX),S)
        lml2=lmm.nLLeval(ldelta,_UY,_UX,S)
        assert SP.absolute(lml1-lml2)<1E-10, 'outch'
        print "lml: %.2f delta lml (rel) : %.2f " % (lml1,(lml1-lml0)/SP.absolute(lml1))
       

if 0:
    covariates = SP.ones([X.shape[0],1])

    t0=time.time()
    LOD0 = lmm.train_associations(X,Y,K,covariates)
    print "t1"
    t1=time.time()
    LOD1 = lmm.train_interactions(X,Y,K,X[:,0:1],covariates,refit_delta0_snp=False)
    print "t2"
    t2=time.time()
    LOD2 = lmm.train_interactions(X,Y,K,X[:,0:10],covariates,refit_delta0_snp=False)
    print "t3"
    t3=time.time()
    LOD2 = lmm.train_interactions(X,Y,K,X[:,0:100],covariates,refit_delta0_snp=False)
    t4=time.time()
    

if 1:
    print "testing train method on all data"
    covariates = SP.ones([X.shape[0],1])
    LOD0 = lmm.train_associations(X,Y,K,covariates)

    kw_args = {'numintervalsAlt':0,'ldeltaminAlt':-1.0,'ldeltamaxAlt':1.0,'numintervals0':100,'ldeltamin0':-10.0,'ldeltamax0':10.0}
    t0 = time.time()
    LOD1 = lmm.train_associations(X,Y,K,covariates,**kw_args)[0]
    t1 = time.time()
    #compare the whole thing
    [LL0, LL, pval, ldelta0, sigg20, beta0, ldelta, sigg2, beta] = lmm_fast.train(X,Y,K,covariates,addBiasTerm=False,**kw_args)
    LOD2 = LL-LL0
    t2 = time.time()
    print SP.absolute(LOD1-LOD2).max()




try:
    import rpy2.robjects.numpy2ri
    from rpy2.robjects import r as r
    from rpy2 import rinterface
except ImportError:
    pass

import numpy as np
import scipy as sp
import sys, pickle, pdb
import scipy.stats as st
import scipy.interpolate
import logging as LG

def R_qvalues(pv):
    rinterface.set_writeconsole(None)
    r.library("qvalue")
    bench = r("qvalue")
    
    results = bench(pv)
    pi0 = results[1]
    qv = np.array(results[2]) 

    return qv, pi0 


def qvalues1(PV,m=None,pi=1.0):
    """estimate q vlaues from a list of Pvalues
    this algorihm is taken from Storey, significance testing for genomic ...
    m: number of tests, (if not len(PV)), pi: fraction of expected true null (1.0 is a conservative estimate)
    """
          
    S = PV.shape
    PV = PV.flatten()
    if m is None:
        m = len(PV) * 1.0
    else:
        m*=1.0
    lPV = len(PV)
    
    #1. sort pvalues
    PV = PV.squeeze()
    IPV = PV.argsort()
    PV  = PV[IPV]

    #2. estimate lambda
    if pi is None:
        lrange = sp.linspace(0.05,0.95,max(lPV/100.0,10))
        pil    = sp.double((PV[:,sp.newaxis]>lrange).sum(axis=0))/lPV
        pilr   = pil/(1.0-lrange)
        #ok, I think for SNPs this is pretty useless, pi is close to 1!
        pi =1.0
        #if there is something useful in there use the something close to 1
        if pilr[-1]<1.0:
            pi = pilr[-1]
            
    #3. initialise q values
    QV_ = pi * m/lPV* PV
    QV_[-1] = min(QV_[-1],1.0)
    #4. update estimate
    for i in xrange(lPV-2,-1,-1):
        QV_[i] = min(pi*m*PV[i]/(i+1.0),QV_[i+1])
    #5. invert sorting
    QV = sp.zeros_like(PV)
    QV[IPV] = QV_

    QV = QV.reshape(S)
    return QV
    
def qvalues(pv, m = None, return_pi0 = False, lowmem = False, pi0 = None, fix_lambda = None):

    original_shape = pv.shape

    assert(pv.min() >= 0 and pv.max() <= 1)

    pv = pv.ravel() # flattens the array in place, more efficient than flatten() 

    if m == None:
	m = float(len(pv))
    else:
	# the user has supplied an m, let's use it
	m *= 1.0

    # if the number of hypotheses is small, just set pi0 to 1
    if len(pv) < 100:
	pi0 = 1.0
    elif pi0 != None:
	pi0 = pi0
    else:
	# evaluate pi0 for different lambdas
	pi0 = []
	lam = sp.arange(0, 0.90, 0.01)
	counts = sp.array([(pv > i).sum() for i in sp.arange(0, 0.9, 0.01)])
	
	if fix_lambda != None:
	    interv_count = (pv > fix_lambda - 0.01).sum()
	    uniform_sim = sp.array([(pv > fix_lambda-0.01).sum()*(i+1) for i in sp.arange(0, len(sp.arange(0, 0.90, 0.01)))][::-1])
	    counts += uniform_sim
	    
	for l in range(len(lam)):
	    pi0.append(counts[l]/(m*(1-lam[l])))

	pi0 = sp.array(pi0)

	# fit natural cubic spline
	tck = sp.interpolate.splrep(lam, pi0, k = 3)
	pi0 = sp.interpolate.splev(lam[-1], tck)
	if pi0 > 1:
	    LG.warning("got pi0 > 1 (%.3f) while estimating qvalues, setting it to 1" % pi0)
	    pi0 = 1.0

	assert(pi0 >= 0 and pi0 <= 1), "%f" % pi0


    if lowmem:
	# low memory version, only uses 1 pv and 1 qv matrices
	qv = sp.zeros((len(pv),))
	last_pv = pv.argmax()
	qv[last_pv] = (pi0*pv[last_pv]*m)/float(m)
	pv[last_pv] = -sp.inf
	prev_qv = last_pv
	for i in xrange(int(len(pv))-2, -1, -1):
	    cur_max = pv.argmax()
	    qv_i = (pi0*m*pv[cur_max]/float(i+1))
	    pv[cur_max] = -sp.inf
	    qv_i1 = prev_qv
	    qv[cur_max] = min(qv_i, qv_i1)
	    prev_qv = qv[cur_max]

    else:
	p_ordered = sp.argsort(pv)    
	pv = pv[p_ordered]
	# estimate qvalues
# 	qv = pi0*m*pv/(sp.arange(len(pv))+1.0)
	
# 	for i in xrange(int(len(qv))-2, 0, -1):
# 	    qv[i] = min([qv[i], qv[i+1]])


	qv = pi0 * m/len(pv) * pv
	qv[-1] = min(qv[-1],1.0)

	for i in xrange(len(pv)-2, -1, -1):
	    qv[i] = min(pi0*m*pv[i]/(i+1.0), qv[i+1])



	# reorder qvalues
	qv_temp = qv.copy()
	qv = sp.zeros_like(qv)
	qv[p_ordered] = qv_temp



    # reshape qvalues
    qv = qv.reshape(original_shape)

    if return_pi0:
	return qv, pi0
    else:
	return qv


def estimate_lambda(pv):
    """estimate lambda form a set of PV"""
    LOD2 = sp.median(st.chi2.isf(pv,1))
    L = (LOD2/0.456)
    return (L)

def LOD2PV(lods):
    PV = (st.chi2.sf(2*lods, 1))
    return PV
        
if __name__ == '__main__':
    qv1, pvalues = pickle.load(open("../experiments_limmi/simulation/results_001_nocorr_0/PANAMA.pickle", "r"))
    # pvalues = pvalues[:20, :100]
    qv = qvalues(pvalues, lowmem = True)
    Rqv, Rpi0 = R_qvalues(pvalues)

    print np.sqrt(sp.power((qv - Rqv), 2).mean())




import numpy as np

def k_fold_cross_validation(X, K, randomise = False):
	"""
	Generates K (training, validation) pairs from the items in X.

	Each pair is a partition of X, where validation is an iterable
	of length len(X)/K. So each training iterable is of length (K-1)*len(X)/K.

	If randomise is true, a copy of X is shuffled before partitioning,
	otherwise its order is preserved in training and validation.
	"""
	if randomise: from random import shuffle; X=list(X); shuffle(X)
	for k in xrange(K):
		training = [x for i, x in enumerate(X) if i % K != k]
		validation = [x for i, x in enumerate(X) if i % K == k]
		yield training, validation

def RMSE(Y, Y_pred):
    return np.sqrt(np.power(Y - Y_pred, 2).mean())



import numpy as np
import scipy as SP
import pdb, sys, pickle
import scipy.linalg
import scipy.linalg as LA
from multiprocessing import Process, Pool
import logging as LG
import linear_regression
sys.path.append(".")
import os
#from IPython.kernel import client

def pinv(x,regulariser=None):
    if regulariser==None:
	return SP.linalg.pinv(x)
    else:
        return (SP.dot(SP.linalg.inv(SP.dot(x.T,x)  + regulariser*SP.eye(x.shape[1])),x.T))

def mlreg(X,Z,missing=None, regulariser = None):
    """ Maximum Likelihood REGression method that is widely used by other classes"""
    if missing is None: missing = SP.zeros(X.shape[0], bool)
    phiI = pinv(X[~missing,:], regulariser = regulariser) 
    return SP.dot(phiI,Z[~missing]).T


def mlpred(X,Z,missing=None, return_W=False):
    if missing is None: missing = SP.zeros(X.shape[0], bool)
    wml = mlreg(X,Z,missing)
    if return_W: return (SP.dot(X[~missing], wml.T), wml)
    else: return SP.dot(X[~missing], wml.T)


def lod(X,Z, miss=None, returnPred=False, covs=None):
    if miss is None: miss = SP.zeros(X.shape[0], bool)
    X = X[~miss]
    Z = Z[~miss]
    Z1 = Z2 = None

    # compute residuals and variances for the nested models
    if covs is None:
        Z1 = Z
        Z2 = Z1 - mlpred(X, Z1)
    else:
        covs = covs[~miss]
        Z1 = Z - mlpred(covs, Z)
        Z2 = Z - mlpred(SP.concatenate((X.T, covs.T)).T, Z)

    V1 = Z1.var(axis=0)
    V2 = Z2.var(axis=0)

    # compute log-odds for both
    LLbg = -X.shape[0]/2*(SP.log(V1*(2*SP.pi))) - 1/(2*V1)*((Z1**2).sum(axis=0))
    LL   = -X.shape[0]/2*(SP.log(V2*(2*SP.pi))) - 1/(2*V2)*((Z2**2).sum(axis=0))

    if not returnPred: return LL - LLbg
    else: return (LL-LLbg, Z - Z2)


def lod2fdr(lod):
    """calculate fdr estimate from likelihood ratio scores
    somewhat a hack but works ok"""
    score = -SP.log(1 + SP.exp(lod))
    return SP.exp(score)
        
def run_associations(Y,X,covs):
    [Ny, N] = Y.shape

    Nx = X.shape[1]

    lods = SP.zeros([Nx, N])

    #add mean column:
    if covs is None: covs = SP.ones([Ny,1])

    for a in xrange(Nx):
        lods[a,:] = lod(X[:,a:a+1],Y,covs=covs)

    return lods

def run_interact(Y, intA, intB, covs,max=False):
    """ Calculate log-odds score for the nested model of including a multiplicative term between intA and intB into the additive model """
    [Ny, N] = Y.shape

    Na = intA.shape[1] # number of interaction terms 1
    Nb = intB.shape[1] # number of interaction terms 2

    # for each snp/gene/factor combination, run a lod
    # snps need to be diced bc of missing values - iterate over them, else in arrays
    lods = SP.zeros([Na, Nb, N])

    #add mean column:
    if covs is None: covs = SP.ones([Ny,1])

    # for each pair of interacting terms
    for a in range(Na):
        for b in range(Nb):
            # calculate additive and interaction terms
            C = SP.concatenate((covs.T, intA[:,a:a+1].T, intB[:,b:b+1].T)).T
            X = intA[:,a:a+1]*intB[:,b:b+1]
            lods[a,b] = lod(X, Y[:,:], covs=C[:,:])        
    if max:
        lods = lods.max(axis=2)
    return lods

    
def run_interact_parallel(Y, intA, intB, covs,max=False,Njobs=5):
    """ Calculate log-odds score for the nested model of including a multiplicative term between intA and intB into the additive model """
    [Ny, N] = Y.shape

    Np = Y.shape[1]    # number of phenotypes
    Na = intA.shape[1] # number of interaction terms 1
    Nb = intB.shape[1] # number of interaction terms 2

    # Ipython conf
    tc = client.TaskClient()
    mec = client.MultiEngineClient()
    eng_ids = mec.get_ids()
    assert eng_ids > 0, "No engine started"	
    assert eng_ids > Njobs, "Not enough engines for the number of jobs required"

    #linear_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "linear_regression.py")
    lmm_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "lmm/lmm_fast.py")
    mec.run(lmm_file)
    #mec.run(linear_file)

    # for each snp/gene/factor combination, run a lod
    # snps need to be diced bc of missing values - iterate over them, else in arrays
    if max:
        lods = SP.zeros([Na, Nb])
    else:
        lods = SP.zeros([Na, Nb, N])
        
    Isplit = SP.array(SP.linspace(0,Na,Njobs+1),dtype='int')

    jobs = []
    for n in xrange(Njobs):
        i0 = Isplit[n]
        i1 = Isplit[n+1]
        intA_= intA[:,i0:i1]
	full_args = (Y, intA_, intB, covs, max)
	job = client.MapTask(run_interact, args=full_args)
	jobs.append(tc.run(job))
	
    tc.barrier(jobs)
    
    #slot results back in:
    for n in xrange(Njobs):
        i0 = Isplit[n]
        i1 = Isplit[n+1]
        lods[i0:i1] = tc.get_task_result(jobs[n])
    
    return lods



"""Covariance functions suitable for GPLVM"""


import scipy as SP
from pygp.covar import CovarianceFunction
import pdb


class LinearInteractCF(CovarianceFunction):
    """

    linear interaction covariance function This covariance assumes that
    part of the inputs X are SNPS and the remainder are latent
    factors.  Furthermore this covariance function maintains a list
    with interaction candidates, where each interaction has it's own
    weight which is a hyperparameter (one per SNP/factor pair).

    """

    def __init__(self, n_dimensions=1, interact_dimension_indices=None,
		 dimension_indices=None, interact_pairs=[], **kw_args):
        """
        n_dimensions                      : total number of dimensions
        dimension_indices(None)           : dimension indices for interactions (1)
        interact_dimension_indices(None)  : dimension indices for interactions (2)
        default: all dimensions as in n_dimensions
        """       

        # inherited constructor
	super(CovarianceFunction, self).__init__(**kw_args)

        # init dimensions
        self._set_dimension_indices(n_dimensions=n_dimensions,
				    interact_dimension_indices = interact_dimension_indices,
				    dimension_indices = dimension_indices,
				    interact_pairs = interact_pairs)
        

    def _set_dimension_indices(self, n_dimensions = 1, interact_dimension_indices = None,
			       dimension_indices = None, interact_pairs=[]):
        """

	set dimension indices for the two types of input dimensions to interact

	"""

        if dimension_indices is not None:
            self.dimension_indices = dimension_indices
        else:
            self.dimension_indices = SP.arange(0,n_dimensions)
        if interact_dimension_indices is not None:
            self.interact_dimension_indices = interact_dimension_indices
        else:
            self.interact_dimension_indices = SP.arange(0,n_dimensions)
        self.interact_pairs = interact_pairs

        # number of hyperparameters:
        self.n_hyperparameters = len(self.interact_pairs)
        
        
    def _filter_i(self,i):
        """

	Filter ineraction dimensions

	"""

        return i[:,self.interact_dimension_indices]

    def _filter_input_dimensions_interact(self,x1,x2):
        """
        """
        if x2 == None:
            xf = self._filter_i(x1)
            return xf,xf
        return self._filter_i(x1), self._filter_i(x2)
    
    def K(self,logtheta,x1,x2=None):
        #1. get inputs data
        if x2 is None:
            x2 = x1
        n = x1.shape[0]

        #2. exponentiate amplitude parameters
        A = SP.exp(2*logtheta)
        RV = SP.zeros([n,n])
        for ip in xrange(len(self.interact_pairs)):
            p = self.interact_pairs[ip]
            i_f = self.dimension_indices[p[0]]
            i_i = self.interact_dimension_indices[p[1]]
            #binarize the SNP state to 0/1 to have a propper interaction again
            x1_f = x1[:,i_f]
            x1_i = x1[:,i_i]
            x2_f = x2[:,i_f]
            x2_i = x2[:,i_i]
            x1_ = x1_f * x1_i
            x2_ = x2_f * x2_i
            RV+= A[ip]*SP.outer(x1_,x2_)
        return RV
        pass

    def Kgrad_theta(self,logtheta,x1,i):
        A = SP.exp(2*logtheta[i])
        p = self.interact_pairs[i]
        i_f = self.dimension_indices[p[0]]
        i_i = self.interact_dimension_indices[p[1]]
        x1_f = x1[:,i_f]
        x1_i = x1[:,i_i]
        x1_ = x1_f * x1_i
        RV= 2*A*SP.outer(x1_,x1_)
        return RV

    def Kgrad_x(self,logtheta,x1,x2,d):
        A = SP.exp(2*logtheta)
        RV = SP.zeros([x1.shape[0],x2.shape[0]])

        #get interaction partners
        n = x1.shape[0]
        for ip in xrange(len(self.interact_pairs)):
            p = self.interact_pairs[ip]
            i_f = self.dimension_indices[p[0]]
            i_i = self.interact_dimension_indices[p[1]]
            if i_f!=d:
                continue
            x1_f = x1[:,i_f]
            x1_i = x1[:,i_i]
            x2_f = x2[:,i_f]
            x2_i = x2[:,i_i]

            x2_ = (x2_f*x2_i)[SP.newaxis,:]
            dx1_ = (x1_i)[:,SP.newaxis]
            RV[:,:] += A[ip] * (dx1_*x2_)
        return RV
        pass

    def Kgrad_xdiag(self,logtheta,x1,d):
        A = SP.exp(2*logtheta)
        RV = SP.zeros([x1.shape[0]])

        #get interaction partners
        n = x1.shape[0]
        for ip in xrange(len(self.interact_pairs)):
            p = self.interact_pairs[ip]
            i_f = self.dimension_indices[p[0]]
            i_i = self.interact_dimension_indices[p[1]]
            #i_state = self.interact_dimension_indices[p[2]]
            if i_f!=d:
                continue
            x1_f = x1[:,i_f]
            x1_i = x1[:,i_i]
            x1_ = x1_f*x1_i
            dx1_ = (x1[:,i_i])
            RV[:] += 2*A[ip] * dx1_*x1_
        return RV
        



import sys, os
import matplotlib

try:
    display = os.environ['DISPLAY']
    if sys.platform == "darwin":
	matplotlib.use('MacOSX')
    if display == "screen":
	matplotlib.use('PDF')
# if running from screen/ssh (no -X), switch to PDF backend
except KeyError:
    matplotlib.use('PDF')

# import pylab as PL
import pylab as plt
plparams = {'axes.labelsize': 22,
          'text.fontsize': 20,
          'title.fontsize': 20,
          'legend.fontsize': 20,
          'xtick.labelsize': 22,
          'ytick.labelsize': 22,
          'text.usetex': False}
plt.rcParams.update(plparams)



import scipy.stats as ST
import scipy as SP
import matplotlib.pylab as PL
def qq_plot(pval=None, filename=None, distr='chi2'):
    """
    draws a QQ plot for a set of pvalues
    Input:
    pval     pvalues 1-dimensional double array, values between 0 and 1
    distr    string indicating the type of distribution [chi2]
    """
    if distr=='chi2':
        n_tests=pval.shape[0];
        pnull=(SP.arange(n_tests)+.5)/n_tests;
        qnull=ST.chi2.isf(pnull,1);
        
        qemp=(ST.chi2.isf(SP.sort(pval),1));
        PL.figure();
        PL.plot(qnull,qemp,'.');
        PL.plot(qnull,qnull);
        PL.ylabel('LOD scores')
        PL.xlabel('$\chi^2$ quantiles')
        if filename is not None:
            PL.savefig(filename)
              
        
        
